import os
import json
import subprocess
import shutil
from datetime import datetime
from argparse import ArgumentParser
from ftplib import FTP, error_perm, error_temp
from webdav3.client import Client as WebDavClient
from pod2gen import Podcast, Episode, Media, Category, Person

def load_config():
    dir_path = os.path.dirname(os.path.abspath(__file__))  # Pobiera ścieżkę katalogu, w którym znajduje się skrypt
    config_path = os.path.join(dir_path, 'config.json')    # Łączy tę ścieżkę z nazwą pliku config.json
    with open(config_path, 'r') as file:
        return json.load(file)

def load_stations():
    dir_path = os.path.dirname(os.path.abspath(__file__))  # Pobiera ścieżkę katalogu, w którym znajduje się skrypt
    stations_path = os.path.join(dir_path, 'stations.json')    # Łączy tę ścieżkę z nazwą pliku stations.json
    with open(stations_path, 'r') as file:
        return json.load(file)

def manage_local_space(directory, config):
    local_max_files = config['local_audio_config']['local_max_files']

    # Jeśli max_files jest równy 0, nie usuwaj żadnych plików
    if local_max_files == 0:
        return

    # Pobierz listę plików w katalogu
    files = [os.path.join(directory, f) for f in os.listdir(directory) if os.path.isfile(os.path.join(directory, f))]
    audio_files = [f for f in files if f.endswith('.mp3') or f.endswith('.opus') or f.endswith('.m4a')]

    # Jeśli liczba plików przekracza max_files, usuń najstarsze
    while len(audio_files) > local_max_files:
        oldest_file = min(audio_files, key=os.path.getmtime)
        os.remove(oldest_file)
        print(f"Usunięto najstarszy plik lokalny: {oldest_file}")
        audio_files.remove(oldest_file)

def manage_ftp_space(ftp, folder_path, config):
    ftp_max_files = config['ftp_config']['ftp_max_files']

    # Jeśli max_files jest równy 0, nie usuwaj żadnych plików
    if ftp_max_files == 0:
        return

    ftp.cwd(folder_path)
    files = ftp.nlst()
    audio_files = [file for file in files if file.endswith('.mp3') or file.endswith('.opus') or file.endswith('.m4a')]

    while len(audio_files) > ftp_max_files:
        oldest_file = min(audio_files, key=lambda x: ftp.sendcmd('MDTM ' + x))
        ftp.delete(oldest_file)
        print(f"Usunięto najstarszy plik: {oldest_file}")
        audio_files.remove(oldest_file)

def parse_date_from_filename(filename):
    # Zakładamy, że nazwa pliku ma format: DD_MM_YY_HH_MM_...
    date_part = filename.split('_')[:5]
    return datetime.strptime('_'.join(date_part), '%d_%m_%y_%H_%M')


def rss(ftp_config, remote_dir, station, show_name):
    ftp_host = ftp_config['ftp_hostname']
    ftp_username = ftp_config['ftp_username']
    ftp_password = ftp_config['ftp_password']
    web_base_url = ftp_config['web_base_url']

    podcast = Podcast()
    podcast.name = f"{show_name} - {station}"
    podcast.authors = [Person(f"{station}")]
    podcast.website = os.path.join(web_base_url, station, show_name)
    podcast.description = f"Podcast {show_name} z {station}"
    podcast.language = "pl"
    podcast.category = Category("News")  # Ensure this is correctly defined
    podcast.explicit = False
    podcast.image = os.path.join(web_base_url, station, show_name, 'podcast.jpg')

    files_info = []

    with FTP(ftp_host) as ftp:
        ftp.login(ftp_username, ftp_password)
        ftp.cwd(remote_dir)
        ftp.set_pasv(True)
        ftp.sendcmd("TYPE I")
        try:
            items = list(ftp.mlsd())
            for item in items:
                name, info = item
                if info.get('type') == 'file' and (name.endswith('.mp3') or name.endswith('.m4a')):
                    size = info.get('size')
                    files_info.append({'name': name, 'size': size})
        except Exception as e:
            print(f"Nie można uzyskać informacji o plikach: {e}")

    remote_dir_relative = remote_dir.replace(ftp_config['ftp_base_path'], '').lstrip('/')
    web_dir_url = os.path.join(web_base_url, station, show_name)

    for file_info in files_info:
        date = parse_date_from_filename(file_info['name'])
        pub_date = date.strftime('%a, %d %b %Y %H:%M:%S GMT')

        episode = Episode()
        episode.title = os.path.splitext(file_info['name'])[0]
        episode.media = Media(url=f"{web_dir_url}/{file_info['name']}", size=file_info['size'], type="audio/mpeg")
        episode.publication_date = pub_date
        episode.summary = 'Opis odcinka'
        podcast.episodes.append(episode)

    rss_content = podcast.rss_str()

    local_rss_file = 'feed.xml'
    with open(local_rss_file, 'wb') as f:
        f.write(rss_content.encode('utf-8'))

    with FTP(ftp_host) as ftp:
        ftp.login(ftp_username, ftp_password)
        ftp.cwd(remote_dir)
        with open(local_rss_file, 'rb') as f:
            ftp.storbinary(f'STOR {os.path.basename(local_rss_file)}', f)
    os.remove(local_rss_file)

    print(f"Wygenerowano i przesłano plik RSS: {local_rss_file} do katalogu: {remote_dir}")

def manage_nextcloud_space(client, folder_path, config, station, show_name):
    nextcloud_max_files = config['nextcloud_config']['nextcloud_max_files']
    if nextcloud_max_files == 0:
        return

    # Użyj argumentów station i show_name zamiast tworzyć ścieżkę wewnątrz funkcji
    full_folder_path = folder_path  # Ścieżka jest już przekazywana jako argument
    # Pobranie listy plików z folderu
    files = client.list(full_folder_path)
    # Filtrowanie tylko plików .mp3, .opus, .m4a
    audio_files = [file for file in files if file.endswith('.mp3') or file.endswith('.opus') or file.endswith('.m4a')]
    # Budowanie pełnych ścieżek dla filtrowanych plików
    full_files_paths = [os.path.join(full_folder_path, file) for file in audio_files]
    # Pobieranie informacji o plikach (może wymagać dostosowania w zależności od API klienta WebDAV)
    files_info = [(file, client.info(file)) for file in full_files_paths]
    # Sortowanie plików po dacie modyfikacji i usuwanie najstarszych, jeśli przekraczają limit
    if len(files_info) > nextcloud_max_files:
        files_sorted = sorted(files_info, key=lambda x: x[1].get('modified'))
        for file_path, _ in files_sorted[:-nextcloud_max_files]:
            # Użycie metody clean do usunięcia pliku
            client.clean(file_path)
            print(f"Usunięto najstarszy plik w Nextcloud: {file_path}")

def upload_ftp(filename, config, station, show_name):
    attempts = 0
    success = False
    while not success and attempts < 5:  # Próbuje ponownie do 5 razy
        try:
            with FTP(config['ftp_config']['ftp_hostname']) as ftp:
                ftp.login(config['ftp_config']['ftp_username'], config['ftp_config']['ftp_password'])
                remote_dir = f"{config['ftp_config']['ftp_base_path']}/{station}/{show_name}"

                # Tworzenie każdego katalogu w ścieżce, jeśli nie istnieje
                parts = remote_dir.split('/')
                current_dir = ''
                for part in parts:
                    if part:
                        current_dir += f'/{part}'
                        try:
                            ftp.cwd(current_dir)
                        except error_perm as e:
                            if 'No such file or directory' in str(e):
                                ftp.mkd(current_dir)
                                ftp.cwd(current_dir)
                            else:
                                raise  # Przekazuje wyjątek, jeśli to inny błąd niż brak katalogu

                # Zapis pliku na serwerze
                with open(filename, 'rb') as file:
                    ftp.storbinary('STOR ' + os.path.basename(filename), file)
                success = True  # Ustaw flagę sukcesu na True, jeśli wszystko przebiegło pomyślnie
                manage_ftp_space(ftp, remote_dir, config)
                rss(config['ftp_config'], remote_dir, station, show_name) #Wywołaj funkcję RSS
                print(f"File {filename} successfully uploaded to FTP.")
        except (error_temp, BrokenPipeError) as e:
            attempts += 1
            print(f"Encountered an error: {e}. Attempting to reconnect and retry... ({attempts})")
        except Exception as e:
            print(f"An unexpected error occurred: {e}")
            break  # Przerwij pętlę, jeśli wystąpi inny nieoczekiwany błąd

    if not success:
        print("Failed to upload file after several attempts.")


def ensure_remote_path(client, remote_path):
    parts = remote_path.strip('/').split('/')
    path = ''
    for part in parts[:-1]:
        path += f"/{part}"
        if not client.check(path):
            client.mkdir(path)

def upload_nextcloud(filename, config, station, show_name):
    options = {
        'webdav_hostname': config['nextcloud_config']['nextcloud_hostname'],  # Poprawiona ścieżka
        'webdav_login': config['nextcloud_config']['nextcloud_login'],  # Poprawiona ścieżka
        'webdav_password': config['nextcloud_config']['nextcloud_password']  # Poprawiona ścieżka
    }
    client = WebDavClient(options)
    remote_dir = f"/public/{station}/{show_name}/"
    remote_path = remote_dir + os.path.basename(filename)

    # Upewnij się, że katalog nadrzędny istnieje
    ensure_remote_path(client, remote_path)

    client.upload_sync(remote_path=remote_path, local_path=filename)
    manage_nextcloud_space(client, remote_dir, config, station, show_name)
    print(f"File {filename} successfully uploaded to NextCloud.")


def record_show(duration, station_url, station, show_name, config):
    # Pobranie ścieżki katalogu, w którym znajduje się bieżący skrypt
    dir_path = os.path.dirname(os.path.abspath(__file__))

    local_audio_config = config['local_audio_config']
    date_str = datetime.now().strftime("%d_%m_%y_%H_%M")

    # Użycie absolutnej ścieżki do katalogu recordings
    output_dir = os.path.join(dir_path, f"recordings/{station}/{show_name}/")
    os.makedirs(output_dir, exist_ok=True)

    output_filename = f"{output_dir}{date_str}_{show_name}_{station}.{local_audio_config['local_audio_format']}"

    # choose codec
    if local_audio_config['local_audio_format'] == 'opus':
        codec = "libopus"
    elif local_audio_config['local_audio_format'] == 'mp3':
        codec = "libmp3lame"
    elif local_audio_config['local_audio_format'] == 'm4a':
        codec = "aac"
    else:
        codec = "copy"  # Domyślny kodek, gdy format nie jest obsługiwany

    ffmpeg_command = [
        "ffmpeg",
        "-t", str(duration * 1),
        "-i", station_url,
        "-acodec", codec,
        "-b:a", f"{local_audio_config['local_audio_bitrate']}k",
        "-ac", str(local_audio_config['local_audio_channels']),
        "-y", output_filename
    ]
    subprocess.run(ffmpeg_command)
    return output_filename


def convert_audio(source_file, config, destination):
    # Pobieranie konfiguracji w zależności od docelowego miejsca przesyłania
    target_config = config['ftp_config'] if destination == 'ftp' else config['nextcloud_config']

    # Pobieranie parametrów docelowych
    target_format = target_config['ftp_audio_format'] if destination == 'ftp' else target_config['nextcloud_audio_format']
    target_bitrate = target_config['ftp_audio_bitrate'] if destination == 'ftp' else target_config['nextcloud_audio_bitrate']
    channels = target_config['ftp_audio_channels'] if destination == 'ftp' else target_config['nextcloud_audio_channels']

    # Określenie nazwy pliku docelowego
    target_filename = os.path.splitext(source_file)[0] + f'_{destination}.{target_format}'

    # Sprawdzanie, czy konwersja jest potrzebna
    if not (config['local_audio_config']['local_audio_format'] == target_format and
            config['local_audio_config']['local_audio_bitrate'] == target_bitrate and
            config['local_audio_config']['local_audio_channels'] == channels):
        print(f"Konwertowanie pliku do formatu {target_format} z bitrate: {target_bitrate}k i {channels} kanałów.")

        if target_format == 'opus':
            codec = "libopus"
            format_output = "opus"
        elif target_format == 'mp3':
            codec = "libmp3lame"
            format_output = "mp3"
        elif target_format == 'm4a':
            codec = "aac"
            format_output = "ipod"  # Używamy 'ipod' aby zapewnić kompatybilność z formatem M4A
        else:
            codec = "copy"
            format_output = target_format  # Domyślnie używamy target_format

        ffmpeg_command = [
            "ffmpeg",
            "-i", source_file,
            "-acodec", codec,
            "-b:a", f"{target_bitrate}k",
            "-ac", str(channels),
            "-f", format_output,
            "-y", target_filename
        ]
        subprocess.run(ffmpeg_command, check=True)
    else:
        print("Plik wejściowy jest już w docelowym formacie i jakości. Tworzenie kopii.")
        shutil.copy(source_file, target_filename)

    return target_filename

def main():
    parser = ArgumentParser()
    parser.add_argument("station", type=str, help="Station name")
    parser.add_argument("show_name", type=str, help="Name of the show")
    parser.add_argument("duration", type=int, help="Duration of the recording in minutes")
    parser.add_argument("-s", "--save", nargs='+', choices=['ftp', 'nextcloud', 'local'], help="Save destinations", default=[])
    args = parser.parse_args()

    config = load_config()
    stations = load_stations()
    station_url = stations[args.station]

    original_file = record_show(args.duration, station_url, args.station, args.show_name, config)

    if 'ftp' in args.save:
        ftp_file = convert_audio(original_file, config, 'ftp')
        upload_ftp(ftp_file, config, args.station, args.show_name)
        # Zawsze usuwaj plik po wysłaniu na FTP
        os.remove(ftp_file)

    if 'nextcloud' in args.save:
        nextcloud_file = convert_audio(original_file, config, 'nextcloud')
        upload_nextcloud(nextcloud_file, config, args.station, args.show_name)
        if nextcloud_file != original_file:
            os.remove(nextcloud_file)

    if 'local' in args.save:
        # Zapisano plik lokalnie, więc teraz zarządzamy przestrzenią
        output_dir = os.path.dirname(original_file)
        manage_local_space(output_dir, config)

    if 'local' not in args.save:
        os.remove(original_file)

if __name__ == "__main__":
    main()
